/*
 *      Copyright (C) 2011-2013 Team XBMC
 *      http://xbmc.org
 *
 *  This Program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This Program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with XBMC; see the file COPYING.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 */

#include "system.h"
#include <EGL/egl.h>
#include "EGLNativeTypePandora.h"
#include "utils/log.h"
#include "guilib/gui3d.h"
#if defined(TARGET_PANDORA)
#include <sys/ioctl.h>
#include <fcntl.h>
#include <bc_cat.h>
#endif
#include "utils/StringUtils.h"

#if defined(TARGET_PANDORA)
bool PandoraInitBCCATDriver();
bool PandoraCloseBCCATDriver();
#endif

CEGLNativeTypePandora::CEGLNativeTypePandora()
{
}

CEGLNativeTypePandora::~CEGLNativeTypePandora()
{
} 

bool CEGLNativeTypePandora::CheckCompatibility()
{
#if defined(TARGET_PANDORA)
  return true;
#endif
  return false;
}

void CEGLNativeTypePandora::Initialize()
{
#if defined(TARGET_PANDORA)
    PandoraInitBCCATDriver();
#endif
  return;
}
void CEGLNativeTypePandora::Destroy()
{
#if defined(TARGET_PANDORA)
    PandoraClearBuffers();
    PandoraCloseBCCATDriver();
#endif

  return;
}

bool CEGLNativeTypePandora::CreateNativeDisplay()
{
  m_nativeDisplay = EGL_DEFAULT_DISPLAY;
  return true;
}

bool CEGLNativeTypePandora::CreateNativeWindow()
{
#if defined(TARGET_PANDORA)
  // Pandora driver use FB, we don't have to create a window for that...
  m_nativeWindow = NULL;
  return true;
#else
  return false;
#endif
}  

bool CEGLNativeTypePandora::GetNativeDisplay(XBNativeDisplayType **nativeDisplay) const
{
  if (!nativeDisplay)
    return false;
  *nativeDisplay = (XBNativeDisplayType*) &m_nativeDisplay;
  return true;
}

bool CEGLNativeTypePandora::GetNativeWindow(XBNativeWindowType **nativeWindow) const
{
#if defined(TARGET_PANDORA)
  if (!nativeWindow)
    return false;
  *nativeWindow = (XBNativeWindowType*) &m_nativeWindow;
  return true;
#else
  return false;
#endif
}

bool CEGLNativeTypePandora::DestroyNativeDisplay()
{
  return true;
}

bool CEGLNativeTypePandora::DestroyNativeWindow()
{
  return true;
}

bool CEGLNativeTypePandora::GetNativeResolution(RESOLUTION_INFO *res) const
{
#if defined(TARGET_PANDORA)

  res->iWidth = 800;
  res->iHeight= 480;

  res->fRefreshRate = 60;
  res->dwFlags= D3DPRESENTFLAG_PROGRESSIVE;
  res->iScreen       = 0;
  res->bFullScreen   = true;
  res->iSubtitles    = (int)(0.965 * res->iHeight);
  res->fPixelRatio   = 1.0f;
  res->iScreenWidth  = res->iWidth;
  res->iScreenHeight = res->iHeight;
  res->strMode       = StringUtils::Format("%dx%d @ %.2f%s - Full Screen", res->iScreenWidth, res->iScreenHeight, res->fRefreshRate, "");
  CLog::Log(LOGNOTICE,"Current resolution: %s\n",res->strMode.c_str());
  return true;
#else
  return false;
#endif
}

bool CEGLNativeTypePandora::SetNativeResolution(const RESOLUTION_INFO &res)
{
  return false;
}

bool CEGLNativeTypePandora::ProbeResolutions(std::vector<RESOLUTION_INFO> &resolutions)
{
  RESOLUTION_INFO res;
  bool ret = false;
  ret = GetNativeResolution(&res);
  if (ret && res.iWidth > 1 && res.iHeight > 1)
  {
    resolutions.push_back(res);
    return true;
  }
  return false;
}

bool CEGLNativeTypePandora::GetPreferredResolution(RESOLUTION_INFO *res) const
{
  return false;
}

bool CEGLNativeTypePandora::ShowWindow(bool show)
{
  return false;
}

#if defined(TARGET_PANDORA)
//---------------------------------------------
// Pandora BC-CAT wrapper driver for streaming textures
//
//---------------------------------------------

// to use streaming texture, we need a handle to bc_cat driver...
static int  bcfd = -1;
static bool pandoraBufferAlloced = false;
static bool pandoraBufferDirty = false;
static int pandoraBufferW = 128;
static int pandoraBufferH = 128;
static int pandoraBufferSize = 128*128*2;
static int pandoraBufferN = 3;
static void** pandoraBufferVAddr = NULL;
static unsigned long * pandoraBufferPAddr = NULL;

bool PandoraInitBCCATDriver()
{
//    BCIO_package ioctl_var;
// initialise bc_cat driver
    if ((bcfd = open("/dev/bc_cat", O_RDWR|O_NDELAY)) == -1) {
        printf("PANDORA: open %s failed\n", "/dev/bc_cat");
        return;
    }
    printf("PANDORA: bc_cat driver initialized succefully\n");
}

bool PandoraCloseBCCATDriver()
{
//    BCIO_package ioctl_var;
// initialise bc_cat driver
    if (bcfd != -1) {
        close(bcfd);
        bcfd = -1;
    }
    
}

bool PandoraSetBufferCount(int buffercount)
{
    // define number of buffer used (default is 3)
    if (pandoraBufferN != buffercount) {
        pandoraBufferN = buffercount;
        pandoraBufferDirty = true;
    }
    return true;
}

bool PandoraSetBufferSize(int width, int height)
{
    // check size, must be multiple of 8 !!!
    if ((width<0) || (width%8!=0)) {
        printf("PANDORA: Error, width must be multiple of 8 !\n");
        return false;
    }
     // define size of buffer used (default is 128x128)
     if ((pandoraBufferW != width) || (pandoraBufferH != height)) {
         pandoraBufferW = widht;
         pandoraBufferH = height;
         pandoraBufferSize = width*height*2;
         pandoraBufferDirty = true;
     }
     return true;
}

bool PandoraInitBuffers()
{
    BCIO_package ioctl_var;
    bc_buf_params_t buf_param;
    // alloc buffer (clear buffers if needed)
    if (pandoraBufferAlloced && !pandoraBufferDirty)
        return true;     // nothing to do :)
    if (pandoraBufferAlloced)
        PandoraClearBuffers();
    if (pandoraBufferPAddr != NULL) 
        free(pandoraBufferPAddr);
    pandoraBufferPAddr = (unsigned long *)malloc(sizeof(unsigned long)*pandoraBufferN);
    if (pandoraBufferVAddr != NULL)
        free(pandoraBufferVAddr;
    pandoraBufferVAddr = malloc(sizeof(void*)*pandoraBufferN);
    if (bcfd == -1)
        return false;
        
    buf_param.count = pandoraBufferN;
    buf_param.width = pandoraBufferW;
    buf_param.height = pandoraBufferH;
    buf_param.type = BC_MEMORY_MMAP;
    buf_param.fourcc = BC_PIX_FMT_UYVY;
    
    if (ioctl(bcfd, BCIOREQ_BUFFERS, &buf_param;) != 0) {
        printf("PANDORA: BCIOREQ_BUFFERS failed\n");
        return false;
    }
    if (ioctl(bcfd, BCIOGET_BUFFERCOUNT, &ioctl_var) != 0) {
        printf("PANDORA: cannot get texture buffer info\n");
        return false;
    }
    if (ioctl_var.output == 0) {
        printf("PANDORA: no texture buffer available\n");
        return false;
    }
    // now match V/P Address
    for (idx = 0; idx < pandoraBufferN; idx++) {
        ioctl_var.input = idx;

        if (ioctl(bcfd, BCIOGET_BUFFERPHYADDR, &ioctl_var) != 0) {
            printf("PANDORA: BCIOGET_BUFFERADDR failed\n");
            return false;
        }

        pandoraBufferPAddr[idx] = ioctl_var.output;
        pandoraBufferVAddr[idx] = mmap(NULL, pandoraBufferSize,
                          PROT_READ | PROT_WRITE, MAP_SHARED,
                          bcfd, buf_paddr[idx]);

        if (pandoraBufferVAddr[idx] == MAP_FAILED) {
            printf("PANDORA: mmap failed\n");
            return false;
        }
    }
    pandoraBufferDirty = false;
    pandoraBufferAlloced = true;
    
    return true;
}

bool PandoraIsBuffersAlloced() {
    // true is buffer has been alloced, false if not (or if some PandoraSetBufferxxxxx called and no InitBuffers)
    return pandoraBufferAlloced;
}

bool PandoraClearBuffers()
{
    // unalloc buffers (if alloced)
    if (!pandoraBufferAlloced)
        return true;
    // free vmap
    for (idx = 0; idx < pandoraBufferN; idx++) {
            if (pandoraBufferVAddr[idx] != MAP_FAILED)
                munmap(pandoraBufferVAddr[idx], pandoraBufferSize);
                
    // free bc_cat buffer
    // no real mecanism in fact for that... will be free at next alloc or when quitting...
    
    // finished
    pandoraBufferDirty = false;
    pandoraBufferAlloced = false;
    return true;
}

unsigned long PandoraGetPAddress(int index) 
{
    // Get Physical address of a buffer (NULL if error or not alloced)
    if (!pandoraBufferAlloced)
        return 0;
    if (!pandoraBufferPAddr)
        return 0;
    return pandoraBufferPAddr[index];
}

void* PandoraGetVAddress(int index)
{
    // Get Virtual address of a buffer (NULL if error or not alloced)
    if (!pandoraBufferAlloced)
        return 0;
    if (!pandoraBufferVAddr)
        return 0;
    return pandoraBufferVAddr[index];
}
#endif
